//------------------------------ $Keywords ----------------------------------
// SelfImage - Disk image manipulation program
// SelfImage_TImageStore.cpp - TImageStore class
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of SelfImage.
//
// SelfImage is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License (Version 2) as
// published by the Free Software Foundation.
//
// SelfImage is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// VCS: $Version: 1 $ $Revision: 6 $
/*
$History: **** V 0.1 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-11-07 2:39:39 AM - 8573 Bytes
$History: * selfimage_timagestore.h - 2005-10-21 10:56:46 AM - 2239 Bytes
$History: * Initial check-in
$History: **** V 0.2 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-11-12 5:22:11 PM - 8837 Bytes
$History: * selfimage_timagestore.h - 2005-11-12 5:19:26 PM - 2243 Bytes
$History: * Typo in copyright area + change in program description
$History: **** V 1.0 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-11-27 6:09:23 AM - 12796 Bytes
$History: * selfimage_timagestore.h - 2005-11-26 6:11:50 PM - 3073 Bytes
$History: * Add decompression support
$History: **** V 1.1 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-12-04 10:52:30 AM - 16143 Bytes
$History: * selfimage_timagestore.h - 2005-12-03 10:33:20 PM - 4433 Bytes
$History: * Add support for determining if a particular area is 
$History: * marked by NTFS as being unused
$History: **** V 1.2 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-12-04 11:03:21 AM - 17164 Bytes
$History: * selfimage_timagestore.h - 2005-12-04 10:54:50 AM - 4447 Bytes
$History: * Add better checking for non NTFS/FAT32 partitions, 
$History: * and document a little better
$History: **** V 1.3 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-12-08 4:12:45 PM - 18840 Bytes
$History: * selfimage_timagestore.h - 2005-12-06 10:25:20 PM - 5075 Bytes
$History: * Network Block Device support
$History: **** V 1.4 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-12-08 4:56:41 PM - 19050 Bytes
$History: * selfimage_timagestore.h - 2005-12-06 10:25:20 PM - 5075 Bytes
$History: * Fix NBD EOF condition
$History: **** V 1.5 by kfitzner ****
$History: * selfimage_timagestore.cpp - 2005-12-09 2:20:38 PM - 19598 Bytes
$History: * selfimage_timagestore.h - 2005-12-08 7:24:02 PM - 4946 Bytes
$History: * Fix FPosition not being set on writes, add configurable 
$History: * timeout and transaction size support
$History: **** Latest ** V 1.6 by kfitzner ** 2005-12-09 5:01:55 PM ****
$History: * Forgot to multiply the transaction size by 1024 - whoops
*/
//----------------------------  $NoKeywords ---------------------------------

//---------------------------------------------------------------------------
// File Notes:
//---------------------------------------------------------------------------
// 12 Sep 2005 - Kurt Fitzner <kfitzner@excelcia.org>
//
// An image "store" represtents any place where, no surprise, an image can
// be stored or retrieved from.  A file, a partition, even a network source.
// This class abstracts opens, closes, reads, etc so that the other classes
// don't need to know (or care) where their data is coming from or going.
//
// This class doesn't need to be particularly thread-safe, since it's a
// little unlikely that, say, the read thread is going to want to try and
// access the write store.
//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop

#include "SelfImage_Utility.h"
#include "SelfImage_TImageStore.h"
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TImageStore constructor that opens the target at the same time
//
__fastcall TImageStore::TImageStore(TImageStoreType isType, AnsiString Target, bool WriteAccess) {
  FType = isType;
  FFileName = Target;
  FCompressionFormat = cfNone;
  bWriteAccess = WriteAccess;
  bIsNTFSFAT32 = false;
  NBDConnection = NULL;
  if (!Open())
    throw ESelfImageError("Could not open store target \"" + Target + "\": " + GetLastErrorMessage());
  FPosition = 0;
  EmptyBlockMap = NULL;
}  // constructor
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TImageStore::~TImageStore()
//
__fastcall TImageStore::~TImageStore() {
  if (EmptyBlockMap)
    delete [] EmptyBlockMap;
  Close();
}  // __fastcall TImageStore::~TImageStore()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Open an image store - this determines the type of the store and whether
// or not NT open functions are available and then decides the best way
// to open the target.
//
bool __fastcall TImageStore::Open(void) {

  switch (Type) {
    case isFile: {
      LARGE_INTEGER nnFileSize;

      DWORD Access     = GENERIC_READ | (bWriteAccess?GENERIC_WRITE:0);
      DWORD ShareMode  = (bWriteAccess?0:FILE_SHARE_READ);
      DWORD CreateMode = (bWriteAccess?CREATE_ALWAYS:OPEN_EXISTING);
      hHandle = CreateFile(FileName.c_str(), Access, ShareMode, NULL, CreateMode, FILE_ATTRIBUTE_NORMAL, NULL);
      if (hHandle != INVALID_HANDLE_VALUE) {
        if (!bWriteAccess) {
          nnFileSize.LowPart = GetFileSize(hHandle, (unsigned long *)&nnFileSize.HighPart);
          FGeometry.Bytes = nnFileSize.QuadPart;
          CheckCompression();
        } else
          FGeometry.Bytes = -1;
      }  // if (hHandle != INVALID_HANDLE_VALUE)
      break;
    }  // case isFile:

    case isDrive: {
      PARTITION_INFORMATION PartitionInfo;
      DISK_GEOMETRY DiskGeometry;
      DWORD Dummy;

      DWORD Access     = GENERIC_READ | (bWriteAccess?GENERIC_WRITE:0);
      DWORD ShareMode  = (bWriteAccess?0:FILE_SHARE_READ|FILE_SHARE_WRITE);
      if (HaveNTCalls)
        hHandle = NTOpen(FileName, SYNCHRONIZE | Access, FILE_ATTRIBUTE_NORMAL, ShareMode, FILE_OPEN, FILE_SYNCHRONOUS_IO_NONALERT | FILE_SEQUENTIAL_ONLY);
      else
        hHandle = CreateFile(FileName.c_str(), GENERIC_READ, ShareMode, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_BACKUP_SEMANTICS, NULL);
      if (hHandle != INVALID_HANDLE_VALUE) {
        DeviceIoControl(hHandle, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &PartitionInfo, sizeof(PartitionInfo), &Dummy, NULL);
        DeviceIoControl(hHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY	, NULL, 0, &DiskGeometry, sizeof(DiskGeometry), &Dummy, NULL);
        if (FileName.Pos("Floppy") || FileName.Pos("A:") || FileName.Pos("B:"))
          FGeometry.Bytes = DiskGeometry.Cylinders.QuadPart * DiskGeometry.TracksPerCylinder * DiskGeometry.SectorsPerTrack * DiskGeometry.BytesPerSector;
        else {
          TBootSector *BootSector = new TBootSector;
          unsigned int BytesRead;
          try {
            FGeometry.Bytes = PartitionInfo.PartitionLength.QuadPart;
            Read(BootSector, sizeof(TBootSector), &BytesRead);
            if (BytesRead == sizeof(TBootSector)) {
              if (!memcmp(BootSector->OEMID, "NTFS    ", 8) || !memcmp(BootSector->OEMID, "MSDOS5.0", 8)) {
                FGeometry.SectorsPerCluster = BootSector->SectorsPerCluster;
                bIsNTFSFAT32 = true;
              }  // if (is NTFS/FAT32)
            }  // if (BytesRead == sizeof(TBootSector))
            Seek(0, FILE_BEGIN);
          }  // try
          __finally {
            delete BootSector;
          }  // __finally
        }  // else if (is not a floppy drive)
        FGeometry.BytesPerSector = DiskGeometry.BytesPerSector;
        FGeometry.Sectors = Geometry.Bytes / Geometry.BytesPerSector;
      }  // if (hHandle != INVALID_HANDLE_VALUE)
      break;
    }  // case isDrive:

    case isNBD: {
      AnsiString Host;
      int nPort = 0;

      SplitConnectionString(FileName, Host, nPort);

      if (NBDConnection)
        delete NBDConnection;
      NBDConnection = new TNBDClient(NULL);
      NBDConnection->Host = Host;
      NBDConnection->Port = nPort;
      NBDConnection->Timeout = SelfImageConfig->Values["NBDTimeout"];
      NBDConnection->TransactionSize = SelfImageConfig->Values["NBDTransactionSize"] * 1024;
      NBDConnection->Open();
      FGeometry.Bytes = NBDConnection->Size;
    }  // case isNBD:
  }  // switch (Type)
  return (hHandle != INVALID_HANDLE_VALUE);
}  // bool __fastcall TImageStore::Open(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Close() - Close an image store
//
bool __fastcall TImageStore::Close(void) {
  if (FType == isNBD && NBDConnection) {
    NBDConnection->Close();
    delete NBDConnection;
    return true;
  }  // if (FType == isNBD && NBDConnection)
  else return CloseHandle(hHandle);
}  // bool __fastcall Close(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Read() - As the name suggests, read from the file/device.  Special care
// has to be taken when reading from handles created with "NtCreateFile()".
// Windows versions based on NT have problems with partial reads of disk
// devices opened with NtCreateFile().  When there isn't enough data in the
// file to satisfy the entire length of the read, ReadFile() is supposed to
// always read as many as there are.  Disk devices of the type
// "\Device\Harddisk0\Partition1" opened with NtCreateFile don't do this.
// When there isn't enough data left to satisfy the entire read, ReadFile()
// will fail entirely and GetLastError() will return ERROR_INVALID_PARAMETER.
// To compensate for this, whenever we encounter ERROR_INVALID_PARAMETER, we
// step down the size of the read to the device's sector size and then read
// until we get an error again.
//
bool __fastcall TImageStore::Read(void *buffer, int nLength, unsigned *nBytesRead) {
  if (FType == isNBD) {
    int ret = NBDConnection->Read(FPosition, buffer, nLength);
    FPosition += ret;
    *nBytesRead = ret;
    return true;
  }  // if (FType == isNBD)
  else {  // if (FType != NBD)
    unsigned nBytesReadTemp, pos;
    bool bSuccess;
    bSuccess = ReadFile(hHandle, buffer, nLength, (unsigned long *)&nBytesReadTemp, NULL);
    if (!bSuccess && GetLastError() == ERROR_INVALID_PARAMETER) {
      // ERROR_INVALID_PARAMETER - step down read size as discussed above
      pos = 0;
      while (ReadFile(hHandle, (char *)buffer + pos, Geometry.BytesPerSector?Geometry.BytesPerSector:512, (unsigned long *)&nBytesReadTemp, NULL))
        pos += nBytesReadTemp;
      nBytesReadTemp = pos;
      bSuccess = true;
    }  // if (!bSuccess)
    *nBytesRead = nBytesReadTemp;
    FPosition += nBytesReadTemp;
    return bSuccess;
  }  // else if (FType != NBD)
}  // bool __fastcall TImageStore::Read(void *buffer, int nLength, unsigned *nBytesRead)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Write() - similar to Read() but different.
//
bool __fastcall TImageStore::Write(void *buffer, int nLength, unsigned *nBytesWritten) {
  bool retval;
  unsigned nWrote;

  if (FType == isNBD)
    retval = (nWrote = NBDConnection->Write(FPosition, buffer, nLength)) != 0;
  else
    retval = WriteFile(hHandle, buffer, nLength, (unsigned long *)&nWrote, NULL);
  FPosition += nWrote;
  *nBytesWritten = nWrote;
  return retval;
}  // bool __fastcall TImageStore::Write(void *buffer, int nLength, unsigned *nBytesWritten)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
//  Set the file position pointer
//
bool __fastcall TImageStore::Seek(unsigned __int64 Offset, DWORD MoveMethod)
{
  if (FType == isNBD) {
    switch (MoveMethod) {
      case FILE_BEGIN: FPosition = Offset; break;
      case FILE_CURRENT: FPosition += Offset; break;
      case FILE_END: FPosition = FGeometry.Bytes + Offset; break;
    }  // switch (MoveMethod)
    return true;
  }  // if (FType == isNBD)
  else {
    LOW_DWORD(Offset) = SetFilePointer(hHandle, LOW_DWORD(Offset), &HIGH_DWORD(Offset), MoveMethod);
    if (GetLastError() == NO_ERROR) {
      FPosition = Offset;
      return true;
    } else
      return false;
  }  // else if (FType != NBD)
}  // bool __fastcall TImageStore::Seek(__int64)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Check to see if the file is compressed.  This is a full check that looks
// at the contents of the file to see if it contains the correct signature
// and also looks to see if it's in SelfImage's block format for parallel
// decompression.
//
void __fastcall TImageStore::CheckCompression(void)
{
  unsigned char *Buffer = NULL;
  AnsiString Extension;

  if (FType != isFile)
    return;

  try {
    Extension = ExtractFileExt(FileName);
    if (Extension == ".gz") {
      unsigned ReadCount;
      Buffer = new unsigned char[20];
      // Read in the gzip header - 20 bytes is all we need
      Read(Buffer, 20, &ReadCount);
      if (ReadCount == 20) {
        // If the file is actually a gzip file then the first two bytes will be 0x1F 0x8B.  The next byte identifies the
        // gzip compression method - since deflate is all we support, we check to make sure it's a deflate gzip file.
        // Deflate is all zlib supports right now - probably all it will ever support.
        if (Buffer[0] == 0x1F && Buffer[1] == 0x8B && Buffer[2] == 0x08)
          // By this point we know for sure it's a gzip file.  Now check to see if it's a gzip that's in SelfImage block format.
          // If it is, it will have a gzip Extra field identified by 'SI' followed by the field length, which must be little-endian 0x0008
          if (Buffer[12] == 'S' && Buffer[13] == 'I' && Buffer[14] == 0x08 && Buffer[15] == 0x00)
            FCompressionFormat = cfGzipBlock;
          else
            FCompressionFormat = cfGzipStream;
      }  // if (ReadCount == 20)
    }  // if (Extension == ".gz")
    else if (Extension == ".bz2") {
      unsigned ReadCount;
      Buffer = new unsigned char[1024 * 1024];
      // Need to read in the entire first bzip2 block, plus the header of the next one.
      // Techinically that's 909610 bytes - make it one meg for good measure
      if (Read(Buffer, 1024 * 1024, &ReadCount)) {
        // If it's a bz2 file, it will start with "BZh" and then have a number from 1 to 9
        if (ReadCount > 20 && Buffer[0] == 'B' && Buffer[1] == 'Z' && Buffer[2] == 'h' && Buffer[3] >= '1' && Buffer[3] <= '9') {
           char *Header = "BZh_1AY&SY";
           Header[3] = Buffer[3];
           // Check to see if there is another full bzip2 header after the first bzip2 block.  If there is,
           // then we have a SelfImage format file - otherwise it's a standard bzip2 file with blocks that
           // we can't tell the size of.
           if (FindString(Buffer+10, ReadCount, Header, true))
             FCompressionFormat = cfBzip2Block;
           else
             FCompressionFormat = cfBzip2Stream;
        }  // if (...is a bzip2 file)
      }  // if (Read(Buffer, 20, &ReadCount))
      // Read in the first compression block
    }  // else if (Extension == ".bz2")
  }  // try
  __finally {
    if (Buffer)
      delete [] Buffer;
    Seek(0, FILE_BEGIN);
  }  // __finally
}  // void __fastcall TImageStore::CheckCompression(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Configure the block map for this device.  The empty block map is an array
// of bool that maps which blocks consist entirely of unallocated clusters.
//
void __fastcall TImageStore::LoadEmptyBlockMap(unsigned int BlockSize)
{
  unsigned nBlockCount, nClustersPerBlock, nBitmapBytes;
  VOLUME_BITMAP_BUFFER *Bitmap;

  // IF this isn't an NTFS or FAT32 partition then we have no business here
  if (!bIsNTFSFAT32 || FType != isDrive)
    return;

  // Determine the number of total blocks on the partition, the number of clusters in a block, and the number of
  // cluster bitmap bytes that are taken by a block.
  nBlockCount = Geometry.Bytes / BlockSize;
  nClustersPerBlock = ceil((double)BlockSize / (double)(Geometry.BytesPerSector * Geometry.SectorsPerCluster));
  nBitmapBytes = ceil(double(nClustersPerBlock) / 8.0);
  // Once we know all that, we can determine the size of the volume bitmap buffer that we need for each block
  Bitmap = (VOLUME_BITMAP_BUFFER *)SysGetMem(sizeof(VOLUME_BITMAP_BUFFER) + nBitmapBytes - 1);

  // This should only ever be called once, but just in case, flush the old empty block map if there is one
  if (EmptyBlockMap)
    delete [] EmptyBlockMap;
  EmptyBlockMap = new bool[nBlockCount];
  memset(EmptyBlockMap, 0, sizeof(bool) * nBlockCount);

  FBlockSize = BlockSize;

  // Iterate through all the blocks and determine if the block resides entirely on clusters that are unallocated
  for (unsigned n = 0; n < nBlockCount; n++) {
    LONGLONG nnStartCluster =  (LONGLONG)n * (LONGLONG)BlockSize / (LONGLONG)(Geometry.BytesPerSector * Geometry.SectorsPerCluster);
    DWORD nBytesReturned;
    bool bIsFree = true;

    // Ask for a cluster bitmap starting with the cluster this block starts on
    int ret = DeviceIoControl(hHandle, FSCTL_GET_VOLUME_BITMAP, &nnStartCluster, sizeof(nnStartCluster), Bitmap, sizeof(VOLUME_BITMAP_BUFFER) + nBitmapBytes, &nBytesReturned, NULL);
    if (!ret && GetLastError() != ERROR_MORE_DATA) {
      memset(EmptyBlockMap, 0, sizeof(bool) * nBlockCount);
      break;
    }  // if (!ret && GetLastError() != ERROR_MORE_DATA)
    // Check to see if all the clusters in this block are free
    for (unsigned b = 0; b < nBitmapBytes; b++)
      if (Bitmap->Buffer[b]) {
        bIsFree = false;
        break;
      }  // sizeof(VOLUME_BITMAP_BUFFER) + nBitmapBytes
    EmptyBlockMap[n] = bIsFree;
  }  // for (unsigned n = 0; n < nnBlockCount; n++)
  SysFreeMem(Bitmap);
  // The last block must always be read normally so that the resulting image is sized properly
  EmptyBlockMap[nBlockCount - 1] = false;
}  // void __fastcall TImageStore:LoadBlockMap(unsigned int BlockSize)
//---------------------------------------------------------------------------

